// Copyright (c) 2009-2025 The Regents of the University of Michigan.
// Part of HOOMD-blue, released under the BSD 3-Clause License.

#pragma once

#include "ExternalPotential.h"

namespace hoomd
    {
namespace hpmc
    {
/** Linear potential based on the distance to a plane.

    Use for e.g. gravity.
*/
class ExternalPotentialLinear : public ExternalPotential
    {
    public:
    ExternalPotentialLinear(std::shared_ptr<SystemDefinition> sysdef)
        : ExternalPotential(sysdef), m_plane_origin(0, 1, 0), m_plane_normal(0, 0, 0),
          m_alpha(sysdef->getParticleData()->getNTypes())
        {
        }
    virtual ~ExternalPotentialLinear() { }

    /// Set the origin of the potential.
    void setPlaneOrigin(const vec3<LongReal>& plane_origin)
        {
        m_plane_origin = plane_origin;
        }

    /// Get the origin.
    vec3<LongReal> getPlaneOrigin() const
        {
        return m_plane_origin;
        }
    /// Set the normal of the potential.
    void setPlaneNormal(const vec3<LongReal>& plane_normal)
        {
        if (dot(plane_normal, plane_normal) == 0)
            {
            throw std::domain_error("plane_normal must have a nonzero magnitude.");
            }
        m_plane_normal = normalize(plane_normal);
        }

    /// Get the normal.
    vec3<LongReal> getPlaneNormal() const
        {
        return m_plane_normal;
        }

    /// Set the linear coefficient
    void setAlpha(const std::string& particle_type, LongReal alpha);

    /// Get the linear coefficient.
    LongReal getAlpha(const std::string& particle_type);

    protected:
    /// A point on the plane.
    vec3<LongReal> m_plane_origin;

    /// The normal vector (unit length) perpendicular to the plane.
    vec3<LongReal> m_plane_normal;

    /// The linear coefficient per particle type.
    std::vector<LongReal> m_alpha;

    /** Implement the evaluation the energy of the external field interacting with one particle.

        @param timestep The current timestep in the simulation
        @param tag_i Tag of the particle
        @param type_i Type index of the particle.
        @param r_i Posiion of the particle in the box.
        @param q_i Orientation of the particle
        @param charge_i Charge of the particle.
        @param trial A value of None indicates that the energy should be evaluated directly.
          Pass Old or New when evaluating the old or new configuration in a trial move.
          Hard potentials always return 0 in old configurations to avoid infinity - infinity.
        @returns Energy of the external interaction (possibly INFINITY).

        Evaluate the linear potential energy.
    */
    virtual LongReal particleEnergyImplementation(uint64_t timestep,
                                                  unsigned int tag_i,
                                                  unsigned int type_i,
                                                  const vec3<LongReal>& r_i,
                                                  const quat<LongReal>& q_i,
                                                  LongReal charge_i,
                                                  Trial trial);
    };

    } // end namespace hpmc

    } // end namespace hoomd
